/* vi: set sw=4 ts=4: */
/*
 * Mini kill/killall implementation for busybox with some changes
 *
 * 1. new "-z" option to send SIGKILL if process does not terminate
 *    before a timeout expires (default timeout is 5 seconds)
 *
 * 2. new "-t N" option to set timeout to N seconds
 *
 * Copyright (C) 1995, 1996 by Bruce Perens <bruce@pixar.com>.
 * Copyright (C) 2001 by Transmeta Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */


#include "busybox.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <ctype.h>
#include <unistd.h>

#ifndef BB_KILL
#error "you must define BB_KILL to use killall"
#endif

#define KILL	0
#define KILLALL	1

struct signal_name {
	const char *name;
	int number;
};

const struct signal_name signames[] = {
	/* POSIX signals */
	{ "HUP",	SIGHUP },	/* 1 */
	{ "INT",	SIGINT }, 	/* 2 */
	{ "QUIT",	SIGQUIT }, 	/* 3 */
	{ "ILL",	SIGILL }, 	/* 4 */
	{ "ABRT",	SIGABRT }, 	/* 6 */
	{ "FPE",	SIGFPE }, 	/* 8 */
	{ "KILL",	SIGKILL }, 	/* 9 */
	{ "SEGV",	SIGSEGV }, 	/* 11 */
	{ "PIPE",	SIGPIPE }, 	/* 13 */
	{ "ALRM",	SIGALRM }, 	/* 14 */
	{ "TERM",	SIGTERM }, 	/* 15 */
	{ "USR1",	SIGUSR1 }, 	/* 10 (arm,i386,m68k,ppc), 30 (alpha,sparc*), 16 (mips) */
	{ "USR2",	SIGUSR2 }, 	/* 12 (arm,i386,m68k,ppc), 31 (alpha,sparc*), 17 (mips) */
	{ "CHLD",	SIGCHLD }, 	/* 17 (arm,i386,m68k,ppc), 20 (alpha,sparc*), 18 (mips) */
	{ "CONT",	SIGCONT }, 	/* 18 (arm,i386,m68k,ppc), 19 (alpha,sparc*), 25 (mips) */
	{ "STOP",	SIGSTOP },	/* 19 (arm,i386,m68k,ppc), 17 (alpha,sparc*), 23 (mips) */
	{ "TSTP",	SIGTSTP },	/* 20 (arm,i386,m68k,ppc), 18 (alpha,sparc*), 24 (mips) */
	{ "TTIN",	SIGTTIN },	/* 21 (arm,i386,m68k,ppc,alpha,sparc*), 26 (mips) */
	{ "TTOU",	SIGTTOU },	/* 22 (arm,i386,m68k,ppc,alpha,sparc*), 27 (mips) */
	/* Miscellaneous other signals */
#ifdef SIGTRAP
	{ "TRAP",	SIGTRAP },	/* 5 */
#endif
#ifdef SIGIOT
	{ "IOT",	SIGIOT }, 	/* 6, same as SIGABRT */
#endif
#ifdef SIGEMT
	{ "EMT",	SIGEMT }, 	/* 7 (mips,alpha,sparc*) */
#endif
#ifdef SIGBUS
	{ "BUS",	SIGBUS },	/* 7 (arm,i386,m68k,ppc), 10 (mips,alpha,sparc*) */
#endif
#ifdef SIGSYS
	{ "SYS",	SIGSYS }, 	/* 12 (mips,alpha,sparc*) */
#endif
#ifdef SIGSTKFLT
	{ "STKFLT",	SIGSTKFLT },	/* 16 (arm,i386,m68k,ppc) */
#endif
#ifdef SIGURG
	{ "URG",	SIGURG },	/* 23 (arm,i386,m68k,ppc), 16 (alpha,sparc*), 21 (mips) */
#endif
#ifdef SIGIO
	{ "IO",		SIGIO },	/* 29 (arm,i386,m68k,ppc), 23 (alpha,sparc*), 22 (mips) */
#endif
#ifdef SIGPOLL
	{ "POLL",	SIGPOLL },	/* same as SIGIO */
#endif
#ifdef SIGCLD
	{ "CLD",	SIGCLD },	/* same as SIGCHLD (mips) */
#endif
#ifdef SIGXCPU
	{ "XCPU",	SIGXCPU },	/* 24 (arm,i386,m68k,ppc,alpha,sparc*), 30 (mips) */
#endif
#ifdef SIGXFSZ
	{ "XFSZ",	SIGXFSZ },	/* 25 (arm,i386,m68k,ppc,alpha,sparc*), 31 (mips) */
#endif
#ifdef SIGVTALRM
	{ "VTALRM",	SIGVTALRM },	/* 26 (arm,i386,m68k,ppc,alpha,sparc*), 28 (mips) */
#endif
#ifdef SIGPROF
	{ "PROF",	SIGPROF },	/* 27 (arm,i386,m68k,ppc,alpha,sparc*), 29 (mips) */
#endif
#ifdef SIGPWR
	{ "PWR",	SIGPWR },	/* 30 (arm,i386,m68k,ppc), 29 (alpha,sparc*), 19 (mips) */
#endif
#ifdef SIGINFO
	{ "INFO",	SIGINFO },	/* 29 (alpha) */
#endif
#ifdef SIGLOST
	{ "LOST",	SIGLOST }, 	/* 29 (arm,i386,m68k,ppc,sparc*) */
#endif
#ifdef SIGWINCH
	{ "WINCH",	SIGWINCH },	/* 28 (arm,i386,m68k,ppc,alpha,sparc*), 20 (mips) */
#endif
#ifdef SIGUNUSED
	{ "UNUSED",	SIGUNUSED },	/* 31 (arm,i386,m68k,ppc) */
#endif
	{0, 0}
};

extern int kill_main(int argc, char **argv)
{
	int whichApp, sig = SIGTERM;
	const char *appUsage;
	int opt_timeout = 5, opt_zap = 0;
	int status = 0;
	pid_t *all_pids = NULL;
	int pids = 0;
	int i;

#ifdef BB_KILLALL
	/* Figure out what we are trying to do here */
	whichApp = (strcmp(applet_name, "killall") == 0)? KILLALL : KILL;
	appUsage = (whichApp == KILLALL)?  killall_usage : kill_usage;
#else
	whichApp = KILL;
	appUsage = kill_usage;
#endif

	argc--;
	argv++;
	/* Parse any options */
	if (argc < 1)
		usage(appUsage);

	while (argc > 0) {
		if (!strcmp("-t", *argv)) {
			if (--argc == 0) {
				usage(appUsage);
			}
			opt_timeout = atoi(*(++argv));
		}
		else if (!strcmp("-z", *argv)) {
			opt_zap = 1;
		}
		else if (!strcmp("-l", *argv)) {
			const struct signal_name *s = signames;
			int col = 0;

			while (s->name != 0) {
				if (col > 60) {
					col = 0;
					putchar('\n');
				}
				else if (col != 0) {
					col++;
					putchar(' ');
				}
				col += printf("%s", (s++)->name);
			}
			putchar('\n');
			exit(0);
		}
		else if (**argv == '-') {
			++(*argv);
			if (isdigit(**argv)) {
				sig = atoi(*argv);
				if (sig < 0 || sig >= NSIG)
					goto end;
			}
			else {
				const struct signal_name *s = signames;

				while (s->name != 0) {
					if (!strcasecmp(s->name, *argv)) {
						sig = s->number;
						break;
					}
					s++;
				}
				if (s->name == 0)
					goto end;
			}
		}
		else {
			/* done parsing options */
			break;
		}
		argv++;
		argc--;
	}

	if (whichApp == KILL) {
		pid_t p = 0;

		while (--argc >= 0) {
			p = strtol(*argv, NULL, 0);
			if (p <= 0) {
				status = 1;
				errorMsg("No such pid %s\n", *argv);
			}
			else {
				pids++;
				all_pids=realloc(all_pids, pids*sizeof(pid_t));
				all_pids[pids - 1] = p;
			}
			argv++;
		}
	}

#if defined (BB_KILLALL)
	else {
		pid_t myself = getpid();
		pid_t *list, *p;
		int found;

		while (--argc >= 0) {
			list = findPidByName(*argv);
			found = 0;
			for(p = list; p && *p != 0; p++) {
				if (*p == myself)
					continue;
				pids++;
				found = 1;
				all_pids=realloc(all_pids, pids*sizeof(pid_t));
				all_pids[pids - 1] = *p;
			}
			if (!found) {
				status = 1;
				errorMsg("%s: No process killed\n", *argv);
			}
			argv++;
		}
	}
#endif /* defined (BB_KILLALL) */

	for(i = 0; i < pids; i++) {
		if (kill(all_pids[i], sig) != 0) {
			status = 1;
			errorMsg("(%d) - %s\n", all_pids[i], strerror(errno));
			all_pids[i] = 0; /* forget this pid */
		}
	}

	if (opt_zap) {
		int left = 0;
		int loops = opt_timeout * 10;

		/* poll for death */
		sig = 0;
		while (loops--) {
			left = 0;
			usleep(100000);
			if (!loops) sig = SIGKILL;	/* zap time! */
			for(i = 0; i < pids; i++) {
				if (all_pids[i]) {
					if (kill(all_pids[i],sig) != 0)
						all_pids[i] = 0;
					else
						left = 1;
				}
			}
			if (!left)
				break;
		}
	}

	exit(status);


  end:
	fatalError("bad signal name: %s\n", *argv);
}

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
